###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.examples.sports_tournament_scheduling.sports_tournament_scheduling_problem import (
    SportsTournamentSchedulingProblem,
)
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import get_statistics

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache PYTHONPATH=. python -m nucs.examples.sports_tournament_scheduling -n 8
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int, default=8)
    parser.add_argument("--symmetry_breaking", action=argparse.BooleanOptionalAction, default=True)
    args = parser.parse_args()
    problem = SportsTournamentSchedulingProblem(args.n, args.symmetry_breaking)
    solver = BacktrackSolver(
        problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
    )
    solution = next(solver.solve())
    print(get_statistics(solver.statistics))
    if solution is not None:
        print(problem.solution_as_matrix(solution))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from nucs.constants import MAX, MIN
from nucs.problems.problem import Problem, is_solved


class TestProblem:
    def test_set_min_value(self) -> None:
        problem = Problem([(0, 2), (0, 2), (0, 2)], [0, 1, 2], [1, 1, 1])
        problem.init()
        problem.set_min_value(1, 2)
        assert problem.shr_domains_arr[1, MIN] == 1
        assert problem.shr_domains_arr[1, MAX] == 2

    def test_is_not_solved_ok(self) -> None:
        problem = Problem([(0, 2), (0, 2), (4, 6)])
        problem.init()
        assert not is_solved(problem.shr_domains_arr)

    def test_is_not_solved_ko(self) -> None:
        problem = Problem([(2, 2), (2, 2), (6, 6)])
        problem.init()
        assert is_solved(problem.shr_domains_arr)

    def test_split_0(self) -> None:
        problem = Problem([(0, 9)])
        problems = problem.split(2, 0)
        assert problems[0].shr_domains_lst[0] == (0, 4)
        assert problems[1].shr_domains_lst[0] == (5, 9)

    def test_split_1(self) -> None:
        problem = Problem([(0, 9)])
        problems = problem.split(3, 0)
        assert problems[0].shr_domains_lst[0] == (0, 3)
        assert problems[1].shr_domains_lst[0] == (4, 6)
        assert problems[2].shr_domains_lst[0] == (7, 9)

    def test_split_2(self) -> None:
        problem = Problem([(0, 11)])
        problems = problem.split(8, 0)
        assert problems[0].shr_domains_lst[0] == (0, 1)
        assert problems[1].shr_domains_lst[0] == (2, 3)
        assert problems[2].shr_domains_lst[0] == (4, 5)
        assert problems[3].shr_domains_lst[0] == (6, 7)
        assert problems[4].shr_domains_lst[0] == (8, 8)
        assert problems[5].shr_domains_lst[0] == (9, 9)
        assert problems[6].shr_domains_lst[0] == (10, 10)
        assert problems[7].shr_domains_lst[0] == (11, 11)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import copy
from typing import Optional, Self, Tuple, Union

import numpy as np
from numba import njit  # type: ignore
from numba.typed import List
from numpy.typing import NDArray

from nucs.constants import END, MAX, MIN, START
from nucs.numpy import (
    new_algorithms,
    new_bounds,
    new_dom_indices,
    new_dom_indices_by_values,
    new_dom_offsets,
    new_dom_offsets_by_values,
    new_not_entailed_propagators,
    new_parameters,
    new_shr_domains_by_values,
    new_shr_domains_propagators,
    new_triggered_propagators,
)
from nucs.propagators.propagators import GET_COMPLEXITY_FCTS, GET_TRIGGERS_FCTS
from nucs.statistics import STATS_IDX_PROBLEM_PROPAGATOR_NB, STATS_IDX_PROBLEM_VARIABLE_NB


class Problem:
    """
    A problem is defined by a list of domains, a list of variables and a list of propagators.
    A domain is computed by adding an offset to a shared domain.
    """

    def __init__(
        self,
        shr_domains_list: List[Union[int, Tuple[int, int]]],
        dom_indices_list: Optional[List[int]] = None,
        dom_offsets_list: Optional[List[int]] = None,
    ):
        """
        Inits the problem.
        :param shr_domains_list: the shared domains expressed as a list
        :param dom_indices_list: the domain indices expressed as a list
        :param dom_offsets_list: the domain offsets expressed as a list
        """
        n = len(shr_domains_list)
        if dom_indices_list is None:
            dom_indices_list = list(range(0, n))
        if dom_offsets_list is None:
            dom_offsets_list = [0] * n
        self.shr_domains_lst = shr_domains_list
        self.dom_indices_lst = dom_indices_list
        self.dom_offsets_lst = dom_offsets_list
        self.propagators: List[Tuple[List[int], int, List[int]]] = []
        self.propagator_nb = 0

    def split(self, split_nb: int, var_idx: int) -> List[Self]:
        """
        Splits a problem into several sub-problems by splitting the domain of a variable.
        :param split_nb: the number of sub-problems
        :param var_idx: the index of the variable
        :return: a list of sub-problems
        """
        shr_dom = self.shr_domains_lst[var_idx]
        if isinstance(shr_dom, int):
            shr_dom_min = shr_dom
            shr_dom_max = shr_dom
        else:
            shr_dom_min = shr_dom[0]
            shr_dom_max = shr_dom[1]
        shr_dom_sz = shr_dom_max - shr_dom_min + 1
        problems = []
        min_idx = shr_dom_min
        for split_idx in range(split_nb):
            problem = copy.deepcopy(self)
            max_idx = min_idx + shr_dom_sz // split_nb - (0 if split_idx < shr_dom_sz % split_nb else 1)
            problem.shr_domains_lst[var_idx] = (min_idx, max_idx)
            min_idx = max_idx + 1
            problems.append(problem)
        return problems

    def split_by_intervals(self, intervals: List[Tuple[int, int]], var_idx: int) -> List[Self]:
        """
        Splits a problem into several sub-problems by assigning intervals to a variable.
        :param intervals: the list of intervals
        :param var_idx: the index of the variable
        :return: a list of sub-problems
        """
        problems = []
        for interval in intervals:
            problem = copy.deepcopy(self)
            problem.shr_domains_lst[var_idx] = interval
            problems.append(problem)
        return problems

    def add_variable(
        self, shr_domain: Union[int, Tuple[int, int]], dom_index: Optional[int] = None, dom_offset: Optional[int] = None
    ) -> None:
        """
        Adds an extra variable to the problem.
        :param shr_domain: the shared domain of the variable
        :param dom_index: the domain index (automatically computed if not defined)
        :param dom_offset: the domain offset (set to 0 if not defined)
        """
        start = len(self.shr_domains_lst)
        if dom_index is None:
            dom_index = start
        if dom_offset is None:
            dom_offset = 0
        self.shr_domains_lst.append(shr_domain)
        self.dom_indices_lst.append(dom_index)
        self.dom_offsets_lst.append(dom_offset)

    def add_variables(
        self,
        shr_domains_list: List[Union[int, Tuple[int, int]]],
        dom_indices_list: Optional[List[int]] = None,
        dom_offsets_list: Optional[List[int]] = None,
    ) -> None:
        """
        Adds extra variabled to the problem.
        :param shr_domains_list: the shared domains of the variables
        :param dom_indices_list: the domain indices (automatically computed if not defined)
        :param dom_offsets_list: the domain offsets (set to 0 if not defined)
        """
        start = len(self.shr_domains_lst)
        n = len(shr_domains_list)
        if dom_indices_list is None:
            dom_indices_list = [start + i for i in range(n)]
        if dom_offsets_list is None:
            dom_offsets_list = [0] * n
        self.shr_domains_lst.extend(shr_domains_list)
        self.dom_indices_lst.extend(dom_indices_list)
        self.dom_offsets_lst.extend(dom_offsets_list)

    def add_propagator(self, propagator: Tuple[List[int], int, List[int]]) -> None:
        """
        Adds an extra propagator.
        :param propagator: the propagator
        """
        self.propagators.append(propagator)

    def add_propagators(self, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
        """
        Adds a list of propagators
        :param propagators: the propagators
        """
        self.propagators.extend(propagators)

    def init(self, statistics: Optional[NDArray] = None) -> None:
        """
        Completes the initialization of the problem by defining the variables and the propagators.
        """
        # Variable and domain initialization
        self.variable_nb = len(self.dom_indices_lst)
        self.shr_domains_arr = new_shr_domains_by_values(self.shr_domains_lst)
        self.dom_indices_arr = new_dom_indices_by_values(self.dom_indices_lst)
        self.dom_offsets_arr = new_dom_offsets_by_values(self.dom_offsets_lst)
        # Sort the propagators based on their estimated amortized complexities.
        self.propagators.sort(key=lambda prop: GET_COMPLEXITY_FCTS[prop[1]](len(prop[0]), prop[2]))
        # Propagator initialization
        self.propagator_nb = len(self.propagators)
        # This is where the triggered propagators will be stored,
        # propagators will be computed in the order of their increasing indices.
        # This is empty at the end of a filter.
        self.triggered_propagators = new_triggered_propagators(self.propagator_nb)
        # This is where the entailed propagators will be stored
        # This is reset in the case of braktrack
        self.not_entailed_propagators = new_not_entailed_propagators(self.propagator_nb)
        self.algorithms = new_algorithms(self.propagator_nb)
        # We will store propagator specific data in a global arrays, we need to compute variables and data bounds.
        self.var_bounds = new_bounds(max(1, self.propagator_nb))  # some redundancy here
        self.param_bounds = new_bounds(max(1, self.propagator_nb))  # some redundancy here
        self.var_bounds[0, START] = self.param_bounds[0, START] = 0
        for prop_idx, prop in enumerate(self.propagators):
            self.algorithms[prop_idx] = prop[1]
            if prop_idx > 0:
                self.var_bounds[prop_idx, START] = self.var_bounds[prop_idx - 1, END]
                self.param_bounds[prop_idx, START] = self.param_bounds[prop_idx - 1, END]
            self.var_bounds[prop_idx, END] = self.var_bounds[prop_idx, START] + len(prop[0])
            self.param_bounds[prop_idx, END] = self.param_bounds[prop_idx, START] + len(prop[2])
        # Bounds have been computed and can now be used. The global arrays are the following:
        self.props_dom_indices = new_dom_indices(self.var_bounds[-1, END])
        self.props_dom_offsets = new_dom_offsets(self.var_bounds[-1, END])
        self.props_parameters = new_parameters(self.param_bounds[-1, END])
        for prop_idx, prop in enumerate(self.propagators):
            prop_vars = prop[0]
            self.props_dom_indices[self.var_bounds[prop_idx, START] : self.var_bounds[prop_idx, END]] = (
                self.dom_indices_arr[prop_vars]
            )  # this is cached for faster access
            self.props_dom_offsets[self.var_bounds[prop_idx, START] : self.var_bounds[prop_idx, END]] = (
                self.dom_offsets_arr[prop_vars]
            )  # this is cached for faster access
            self.props_parameters[self.param_bounds[prop_idx, START] : self.param_bounds[prop_idx, END]] = prop[2]
        self.props_dom_offsets = self.props_dom_offsets.reshape((-1, 1))
        self.shr_domains_propagators = new_shr_domains_propagators(len(self.shr_domains_lst), self.propagator_nb)
        for prop_idx, prop in enumerate(self.propagators):
            triggers = GET_TRIGGERS_FCTS[prop[1]](len(prop[0]), prop[2])
            for prop_var_idx, prop_var in enumerate(prop[0]):
                self.shr_domains_propagators[self.dom_indices_arr[prop_var], :, prop_idx] = triggers[prop_var_idx, :]
        if statistics is not None:
            statistics[STATS_IDX_PROBLEM_PROPAGATOR_NB] = self.propagator_nb
            statistics[STATS_IDX_PROBLEM_VARIABLE_NB] = self.variable_nb

    def reset(self, choice_point: Optional[Tuple[NDArray, NDArray]] = None) -> None:
        if choice_point is None:
            self.shr_domains_arr = new_shr_domains_by_values(self.shr_domains_lst)
            self.not_entailed_propagators.fill(True)
            self.triggered_propagators.fill(True)
        else:
            self.shr_domains_arr, self.not_entailed_propagators = choice_point
            np.copyto(self.triggered_propagators, self.not_entailed_propagators)

    def get_min_value(self, var_idx: int) -> int:
        """
        Gets the minimal value of a variable.
        :param var_idx: the index of the variable
        :return: the minimal value
        """
        return self.shr_domains_arr[self.dom_indices_arr[var_idx], MIN] + self.dom_offsets_arr[var_idx]

    def get_max_value(self, var_idx: int) -> int:
        """
        Gets the maximal value of a variable.
        :param var_idx: the index of the variable
        :return: the maximal value
        """
        return self.shr_domains_arr[self.dom_indices_arr[var_idx], MAX] + self.dom_offsets_arr[var_idx]

    def set_min_value(self, var_idx: int, min_value: int) -> None:
        """
        Sets the minimal value of a variable.
        :param var_idx: the index of the variable
        :param min_value: the minimal value
        """
        self.shr_domains_arr[self.dom_indices_arr[var_idx], MIN] = min_value - self.dom_offsets_arr[var_idx]

    def set_max_value(self, var_idx: int, max_value: int) -> None:
        """
        Sets the maximal value of a variable.
        :param var_idx: the index of the variable
        :param min_value: the maximal value
        """
        self.shr_domains_arr[self.dom_indices_arr[var_idx], MAX] = max_value - self.dom_offsets_arr[var_idx]

    def get_solution(self) -> NDArray:
        """
        Returns the solution to the problem.
        :return: a Numpy array
        """
        return self.shr_domains_arr[self.dom_indices_arr, MIN] + self.dom_offsets_arr

    def __str__(self) -> str:
        return f"domains={self.shr_domains_arr}, indices={self.dom_indices_arr}, offsets={self.dom_offsets_arr}"


@njit(cache=True)
def is_solved(shr_domains: NDArray) -> bool:
    """
    Returns true iff the problem is solved.
    :return: a boolean
    """
    return bool(np.all(np.equal(shr_domains[:, MIN], shr_domains[:, MAX])))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from multiprocessing import Queue
from typing import Callable, Iterator, List, Optional

import numpy as np
from numpy.typing import NDArray

from nucs.constants import PROBLEM_INCONSISTENT, PROBLEM_SOLVED, PROBLEM_TO_FILTER
from nucs.problems.problem import Problem
from nucs.solvers.consistency_algorithms import bound_consistency_algorithm
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic, min_value_dom_heuristic
from nucs.solvers.solver import Solver
from nucs.statistics import (
    STATS_IDX_OPTIMIZER_SOLUTION_NB,
    STATS_IDX_SOLVER_BACKTRACK_NB,
    STATS_IDX_SOLVER_CHOICE_DEPTH,
    STATS_IDX_SOLVER_CHOICE_NB,
    STATS_IDX_SOLVER_SOLUTION_NB,
    init_statistics,
)


def decrease_max(problem: Problem, var_idx: int, value: int) -> None:
    """
    Decreases the max of a variable
    :param problem: the problem
    :param var_idx: the index of the variable
    :param value: the current max
    """
    problem.set_max_value(var_idx, value - 1)


def increase_min(problem: Problem, var_idx: int, value: int) -> None:
    """
    Increases the min of a variable
    :param problem: the problem
    :param var_idx: the index of the variable
    :param value: the current min
    """
    problem.set_min_value(var_idx, value + 1)


class BacktrackSolver(Solver):
    """
    A solver relying on a backtracking mechanism.
    """

    def __init__(
        self,
        problem: Problem,
        consistency_algorithm: Callable = bound_consistency_algorithm,
        var_heuristic: Callable = first_not_instantiated_var_heuristic,
        dom_heuristic: Callable = min_value_dom_heuristic,
        buffer_size: int = 1024,
    ):
        """
        Inits the solver.
        :param problem: the problem
        :param consistency_algorithm: a consistency algorithm (usually bound consistency)
        :param var_heuristic: a heuristic for selecting a variable/domain
        :param dom_heuristic: a heuristic for reducing a domain
        """
        self.problem = problem
        self.statistics = init_statistics()
        self.choice_points = []  # type: ignore
        self.consistency_algorithm = consistency_algorithm
        self.var_heuristic = var_heuristic
        self.dom_heuristic = dom_heuristic
        self.buffer_size = buffer_size

    def init_problem(self) -> None:
        """
        Inits the problem.
        """
        self.problem.init(self.statistics)

    def solve_one(self) -> Optional[NDArray]:
        """
        Find at most one solution.
        :return: the solution if it exists or None
        """
        while (status := self.make_choice()) == PROBLEM_TO_FILTER:
            pass
        return self.problem.get_solution() if status == PROBLEM_SOLVED else None

    def solve(self) -> Iterator[NDArray]:
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        self.init_problem()
        while (solution := self.solve_one()) is not None:
            yield solution
            if not self.backtrack():
                break

    def make_choice(self) -> int:
        """
        Makes a choice and returns a status
        :return: the status as an integer
        """
        # first filter
        while (status := self.consistency_algorithm(self.statistics, self.problem)) == PROBLEM_INCONSISTENT:
            if not self.backtrack():
                return PROBLEM_INCONSISTENT
        if status == PROBLEM_SOLVED:
            self.statistics[STATS_IDX_SOLVER_SOLUTION_NB] += 1
            return PROBLEM_SOLVED
        # then make a choice
        dom_idx = self.var_heuristic(self.problem.shr_domains_arr)
        shr_domains_copy = self.problem.shr_domains_arr.copy(order="F")
        not_entailed_propagators_copy = self.problem.not_entailed_propagators.copy()
        self.choice_points.append((shr_domains_copy, not_entailed_propagators_copy))
        event = self.dom_heuristic(self.problem.shr_domains_arr[dom_idx], shr_domains_copy[dom_idx])
        np.logical_or(
            self.problem.triggered_propagators,
            self.problem.shr_domains_propagators[dom_idx, event],
            self.problem.triggered_propagators,
        )
        self.statistics[STATS_IDX_SOLVER_CHOICE_NB] += 1
        cp_max_depth = len(self.choice_points)
        if cp_max_depth > self.statistics[STATS_IDX_SOLVER_CHOICE_DEPTH]:
            self.statistics[STATS_IDX_SOLVER_CHOICE_DEPTH] = cp_max_depth
        return PROBLEM_TO_FILTER

    def minimize(self, variable_idx: int) -> Optional[NDArray]:
        """
        Return the solution that minimizes a variable.
        :param variable_idx: the index of the variable to minimize
        :return: the optimal solution if it exists or None
        """
        return self.optimize(variable_idx, decrease_max)

    def maximize(self, variable_idx: int) -> Optional[NDArray]:
        """
        Return the solution that maximizes a variable.
        :param variable_idx: the index of the variable to maximize
        :return: the optimal solution if it exists or None
        """
        return self.optimize(variable_idx, increase_min)

    def optimize(self, variable_idx: int, update_target_domain: Callable) -> Optional[NDArray]:
        self.init_problem()
        best_solution = None
        while (solution := self.solve_one()) is not None:
            best_solution = solution
            self.statistics[STATS_IDX_OPTIMIZER_SOLUTION_NB] += 1
            self.reset()
            update_target_domain(self.problem, variable_idx, best_solution[variable_idx])
        return best_solution

    def minimize_and_queue(self, variable_idx: int, processor_idx: int, queue: Queue) -> None:
        """
        Enqueues the solution that minimizes a variable.
        :param variable_idx: the index of the variable to minimize
        :param processor_idx: the index of the processor running the minimizer
        :param queue: the queue
        """
        self.optimize_and_queue(variable_idx, decrease_max, processor_idx, queue)

    def maximize_and_queue(self, variable_idx: int, processor_idx: int, queue: Queue) -> None:
        """
        Enqueues the solution that maximizes a variable.
        :param variable_idx: the index of the variable to maximizer
        :param processor_idx: the index of the processor running the maximizer
        :param queue: the queue
        """
        self.optimize_and_queue(variable_idx, increase_min, processor_idx, queue)

    def add_to_queue(
        self, buffer: List[Optional[NDArray]], queue: Queue, processor_idx: int, solution: Optional[NDArray]
    ) -> None:
        buffer.append(solution)
        if len(buffer) == self.buffer_size or solution is None:
            queue.put((processor_idx, buffer.copy(), self.statistics))
            buffer.clear()

    def solve_and_queue(self, processor_idx: int, queue: Queue) -> None:
        self.init_problem()
        buffer: List[Optional[NDArray]] = []
        while (solution := self.solve_one()) is not None:
            self.add_to_queue(buffer, queue, processor_idx, solution)
            if not self.backtrack():
                break
        self.add_to_queue(buffer, queue, processor_idx, None)

    def optimize_and_queue(
        self, variable_idx: int, update_target_domain: Callable, processor_idx: int, queue: Queue
    ) -> None:
        self.init_problem()
        buffer: List[Optional[NDArray]] = []
        while (solution := self.solve_one()) is not None:
            self.statistics[STATS_IDX_OPTIMIZER_SOLUTION_NB] += 1
            self.add_to_queue(buffer, queue, processor_idx, solution)
            self.reset()
            update_target_domain(self.problem, variable_idx, solution[variable_idx])
        self.add_to_queue(buffer, queue, processor_idx, None)

    def backtrack(self) -> bool:
        """
        Backtracks and updates the problem's domains
        :return: true iff it is possible to backtrack
        """
        if len(self.choice_points) == 0:
            return False
        self.statistics[STATS_IDX_SOLVER_BACKTRACK_NB] += 1
        self.problem.reset(self.choice_points.pop())  # TODO: optimize by reusing
        return True

    def reset(self) -> None:
        """
        Resets the solver by resetting the problem and the choice points.
        """
        self.choice_points.clear()
        self.problem.reset()



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.examples.queens.queens_problem import QueensProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic, smallest_domain_var_heuristic
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver
from nucs.statistics import get_statistics

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache PYTHONPATH=. python -m nucs.examples.queens -n 10
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--processors", type=int, default=1)
    parser.add_argument("--ff", type=bool, action=argparse.BooleanOptionalAction, default=False)
    args = parser.parse_args()
    problem = QueensProblem(args.n)
    problems = problem.split(args.processors, 0)
    solver = MultiprocessingSolver(
        [
            BacktrackSolver(
                problem,
                var_heuristic=smallest_domain_var_heuristic if args.ff else first_not_instantiated_var_heuristic,
            )
            for problem in problems
        ]
    )
    solver.solve_all()
    print(get_statistics(solver.statistics))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from multiprocessing import Process, Queue
from typing import Iterator, List, Optional

from numpy.typing import NDArray

from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.solver import Solver
from nucs.statistics import init_statistics


class MultiprocessingSolver(Solver):
    """
    A solver relying on the multiprocessing package.
    This solver delegates resolution to a set of solvers.
    """

    def __init__(self, solvers: List[BacktrackSolver]):
        self.statistics = [init_statistics() for _ in solvers]
        self.solvers = solvers

    def solve(self) -> Iterator[NDArray]:
        queue: Queue = Queue()
        for processor_idx, solver in enumerate(self.solvers):
            Process(target=solver.solve_and_queue, args=(processor_idx, queue)).start()
        nb = len(self.solvers)
        while nb > 0:
            processor_idx, solutions, statistics = queue.get()
            self.statistics[processor_idx] = statistics
            for solution in solutions:
                if solution is None:
                    nb -= 1
                else:
                    yield solution

    def minimize(self, variable_idx: int) -> Optional[NDArray]:
        queue: Queue = Queue()
        for processor_idx, solver in enumerate(self.solvers):
            Process(target=solver.minimize_and_queue, args=(variable_idx, processor_idx, queue)).start()
        best_solution = None
        nb = len(self.solvers)
        while nb > 0:
            processor_idx, solutions, statistics = queue.get()
            self.statistics[processor_idx] = statistics
            for solution in solutions:
                if solution is None:
                    nb -= 1
                elif best_solution is None or solution[variable_idx] < best_solution[variable_idx]:
                    best_solution = solution
        return best_solution

    def maximize(self, variable_idx: int) -> Optional[NDArray]:
        queue: Queue = Queue()
        for processor_idx, solver in enumerate(self.solvers):
            Process(target=solver.maximize_and_queue, args=(variable_idx, processor_idx, queue)).start()
        best_solution = None
        nb = len(self.solvers)
        while nb > 0:
            processor_idx, solutions, statistics = queue.get()
            self.statistics[processor_idx] = statistics
            for solution in solutions:
                if solution is None:
                    nb -= 1
                elif best_solution is None or solution[variable_idx] > best_solution[variable_idx]:
                    best_solution = solution
        return best_solution



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import pytest

from nucs.examples.magic_sequence.magic_sequence_problem import MagicSequenceProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import last_not_instantiated_var_heuristic, min_value_dom_heuristic
from nucs.statistics import STATS_IDX_SOLVER_SOLUTION_NB


class TestMagicSequence:

    @pytest.mark.parametrize("size,zero_nb", [(50, 46), (100, 96), (200, 196)])
    def test_magic_sequence(self, size: int, zero_nb: int) -> None:
        problem = MagicSequenceProblem(size)
        solver = BacktrackSolver(
            problem, var_heuristic=last_not_instantiated_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solutions = solver.find_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 1
        assert solutions[0][0] == zero_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from nucs.examples.donald.donald_problem import DonaldProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import min_value_dom_heuristic, smallest_domain_var_heuristic
from nucs.statistics import STATS_IDX_SOLVER_SOLUTION_NB


class TestDonald:
    def test_donald(self) -> None:
        problem = DonaldProblem()
        solver = BacktrackSolver(
            problem, var_heuristic=smallest_domain_var_heuristic, dom_heuristic=min_value_dom_heuristic
        )
        solutions = solver.find_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 1
        assert solutions[0].tolist() == [4, 3, 5, 9, 1, 8, 6, 2, 7, 0]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import Callable, Iterator, List, Optional

from numpy._typing import NDArray


class Solver:
    """
    A solver.
    """

    def solve(self) -> Iterator[NDArray]:  # type: ignore
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        pass

    def solve_all(self, func: Optional[Callable] = None) -> None:
        """
        Finds all solutions.
        """
        for solution in self.solve():
            if func is not None:
                func(solution)

    def find_all(self) -> List[NDArray]:
        """
        Finds all solutions.
        """
        solutions = []
        self.solve_all(lambda solution: solutions.append(solution))
        return solutions

    def minimize(self, variable_idx: int) -> Optional[NDArray]:  # type: ignore
        """
        Finds, if it exists, the solution to the problem that minimizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        pass

    def maximize(self, variable_idx: int) -> Optional[NDArray]:  # type: ignore
        """
        Finds, if it exists, the solution to the problem that maximizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        pass



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from rich import print

from nucs.examples.knapsack.knapsack_problem import KnapsackProblem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.heuristics import first_not_instantiated_var_heuristic, max_value_dom_heuristic
from nucs.statistics import get_statistics

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache PYTHONPATH=. python -m nucs.examples.knapsack
if __name__ == "__main__":
    problem = KnapsackProblem(
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        55,
    )
    solver = BacktrackSolver(
        problem, var_heuristic=first_not_instantiated_var_heuristic, dom_heuristic=max_value_dom_heuristic
    )
    solution = solver.maximize(problem.weight)
    print(get_statistics(solver.statistics))
    print(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import Dict, List, Union

import numpy as np
from numpy.typing import NDArray


def init_statistics() -> NDArray:
    """
    Inits a Numpy array for storing the statistics.
    :return: a Numpy array
    """
    return np.array([0] * STATS_MAX, dtype=np.int64)


def sum_stats(stats: List[NDArray], index: int) -> int:
    return sum(int(s[index]) for s in stats)


def max_stats(stats: List[NDArray], index: int) -> int:
    return max(int(s[index]) for s in stats)


def get_statistics(stats: Union[NDArray, List[NDArray]]) -> Dict[str, int]:
    """
    Returns the statistics as a dictionary.
    :param stats: a Numpy array of statistics
    :return: a dictionary
    """
    if isinstance(stats, list):
        return {
            STATS_LBL_OPTIMIZER_SOLUTION_NB: sum_stats(stats, STATS_IDX_OPTIMIZER_SOLUTION_NB),
            STATS_LBL_PROBLEM_FILTER_NB: sum_stats(stats, STATS_IDX_PROBLEM_FILTER_NB),
            STATS_LBL_PROBLEM_PROPAGATOR_NB: int(stats[0][STATS_IDX_PROBLEM_PROPAGATOR_NB]),
            STATS_LBL_PROBLEM_VARIABLE_NB: int(stats[0][STATS_IDX_PROBLEM_VARIABLE_NB]),
            STATS_LBL_PROPAGATOR_ENTAILMENT_NB: sum_stats(stats, STATS_IDX_PROPAGATOR_ENTAILMENT_NB),
            STATS_LBL_PROPAGATOR_FILTER_NB: sum_stats(stats, STATS_IDX_PROPAGATOR_FILTER_NB),
            STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB: sum_stats(stats, STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB),
            STATS_LBL_PROPAGATOR_INCONSISTENCY_NB: sum_stats(stats, STATS_IDX_PROPAGATOR_INCONSISTENCY_NB),
            STATS_LBL_SOLVER_BACKTRACK_NB: sum_stats(stats, STATS_IDX_SOLVER_BACKTRACK_NB),
            STATS_LBL_SOLVER_CHOICE_NB: sum_stats(stats, STATS_IDX_SOLVER_CHOICE_NB),
            STATS_LBL_SOLVER_CHOICE_DEPTH: max_stats(stats, STATS_IDX_SOLVER_CHOICE_DEPTH),
            STATS_LBL_SOLVER_SOLUTION_NB: sum_stats(stats, STATS_IDX_SOLVER_SOLUTION_NB),
        }
    else:
        return {
            STATS_LBL_OPTIMIZER_SOLUTION_NB: int(stats[STATS_IDX_OPTIMIZER_SOLUTION_NB]),
            STATS_LBL_PROBLEM_FILTER_NB: int(stats[STATS_IDX_PROBLEM_FILTER_NB]),
            STATS_LBL_PROBLEM_PROPAGATOR_NB: int(stats[STATS_IDX_PROBLEM_PROPAGATOR_NB]),
            STATS_LBL_PROBLEM_VARIABLE_NB: int(stats[STATS_IDX_PROBLEM_VARIABLE_NB]),
            STATS_LBL_PROPAGATOR_ENTAILMENT_NB: int(stats[STATS_IDX_PROPAGATOR_ENTAILMENT_NB]),
            STATS_LBL_PROPAGATOR_FILTER_NB: int(stats[STATS_IDX_PROPAGATOR_FILTER_NB]),
            STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB: int(stats[STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB]),
            STATS_LBL_PROPAGATOR_INCONSISTENCY_NB: int(stats[STATS_IDX_PROPAGATOR_INCONSISTENCY_NB]),
            STATS_LBL_SOLVER_BACKTRACK_NB: int(stats[STATS_IDX_SOLVER_BACKTRACK_NB]),
            STATS_LBL_SOLVER_CHOICE_NB: int(stats[STATS_IDX_SOLVER_CHOICE_NB]),
            STATS_LBL_SOLVER_CHOICE_DEPTH: int(stats[STATS_IDX_SOLVER_CHOICE_DEPTH]),
            STATS_LBL_SOLVER_SOLUTION_NB: int(stats[STATS_IDX_SOLVER_SOLUTION_NB]),
        }


STATS_MAX = 12
(
    STATS_IDX_OPTIMIZER_SOLUTION_NB,
    STATS_IDX_PROBLEM_FILTER_NB,
    STATS_IDX_PROBLEM_PROPAGATOR_NB,
    STATS_IDX_PROBLEM_VARIABLE_NB,
    STATS_IDX_PROPAGATOR_ENTAILMENT_NB,
    STATS_IDX_PROPAGATOR_FILTER_NB,
    STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_INCONSISTENCY_NB,
    STATS_IDX_SOLVER_BACKTRACK_NB,
    STATS_IDX_SOLVER_CHOICE_NB,
    STATS_IDX_SOLVER_CHOICE_DEPTH,
    STATS_IDX_SOLVER_SOLUTION_NB,
) = tuple(range(STATS_MAX))

STATS_LBL_OPTIMIZER_SOLUTION_NB = "OPTIMIZER_SOLUTION_NB"
STATS_LBL_PROBLEM_FILTER_NB = "PROBLEM_FILTER_NB"
STATS_LBL_PROBLEM_PROPAGATOR_NB = "PROBLEM_PROPAGATOR_NB"
STATS_LBL_PROBLEM_VARIABLE_NB = "PROBLEM_VARIABLE_NB"
STATS_LBL_PROPAGATOR_ENTAILMENT_NB = "PROPAGATOR_ENTAILMENT_NB"
STATS_LBL_PROPAGATOR_FILTER_NB = "PROPAGATOR_FILTER_NB"
STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB = "PROPAGATOR_FILTER_NO_CHANGE_NB"
STATS_LBL_PROPAGATOR_INCONSISTENCY_NB = "PROPAGATOR_INCONSISTENCY_NB"
STATS_LBL_SOLVER_BACKTRACK_NB = "SOLVER_BACKTRACK_NB"
STATS_LBL_SOLVER_CHOICE_NB = "SOLVER_CHOICE_NB"
STATS_LBL_SOLVER_CHOICE_DEPTH = "SOLVER_CHOICE_DEPTH"
STATS_LBL_SOLVER_SOLUTION_NB = "SOLVER_SOLUTION_NB"



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np
from numba import int32, int64, njit, types  # type: ignore
from numpy.typing import NDArray

from nucs.numba import NUMBA_DISABLE_JIT, build_function_address_list
from nucs.propagators.affine_eq_propagator import (
    compute_domains_affine_eq,
    get_complexity_affine_eq,
    get_triggers_affine_eq,
)
from nucs.propagators.affine_geq_propagator import (
    compute_domains_affine_geq,
    get_complexity_affine_geq,
    get_triggers_affine_geq,
)
from nucs.propagators.affine_leq_propagator import (
    compute_domains_affine_leq,
    get_complexity_affine_leq,
    get_triggers_affine_leq,
)
from nucs.propagators.alldifferent_propagator import (
    compute_domains_alldifferent,
    get_complexity_alldifferent,
    get_triggers_alldifferent,
)
from nucs.propagators.and_propagator import compute_domains_and, get_complexity_and, get_triggers_and
from nucs.propagators.count_eq_propagator import (
    compute_domains_count_eq,
    get_complexity_count_eq,
    get_triggers_count_eq,
)
from nucs.propagators.dummy_propagator import compute_domains_dummy, get_complexity_dummy, get_triggers_dummy
from nucs.propagators.element_lic_propagator import (
    compute_domains_element_lic,
    get_complexity_element_lic,
    get_triggers_element_lic,
)
from nucs.propagators.element_liv_propagator import (
    compute_domains_element_liv,
    get_complexity_element_liv,
    get_triggers_element_liv,
)
from nucs.propagators.exactly_eq_propagator import (
    compute_domains_exactly_eq,
    get_complexity_exactly_eq,
    get_triggers_exactly_eq,
)
from nucs.propagators.exactly_true_propagator import (
    compute_domains_exactly_true,
    get_complexity_exactly_true,
    get_triggers_exactly_true,
)
from nucs.propagators.gcc_propagator import compute_domains_gcc, get_complexity_gcc, get_triggers_gcc
from nucs.propagators.lexicographic_leq_propagator import (
    compute_domains_lexicographic_leq,
    get_complexity_lexicographic_leq,
    get_triggers_lexicographic_leq,
)
from nucs.propagators.max_eq_propagator import compute_domains_max_eq, get_complexity_max_eq, get_triggers_max_eq
from nucs.propagators.max_leq_propagator import compute_domains_max_leq, get_complexity_max_leq, get_triggers_max_leq
from nucs.propagators.min_eq_propagator import compute_domains_min_eq, get_complexity_min_eq, get_triggers_min_eq
from nucs.propagators.min_geq_propagator import compute_domains_min_geq, get_complexity_min_geq, get_triggers_min_geq
from nucs.propagators.relation_propagator import (
    compute_domains_relation,
    get_complexity_relation,
    get_triggers_relation,
)

# The ordinals of the algorithms for all propagators (sorted by alphabetical ordering).
(
    ALG_AND,
    ALG_AFFINE_EQ,
    ALG_AFFINE_GEQ,
    ALG_AFFINE_LEQ,
    ALG_ALLDIFFERENT,
    ALG_COUNT_EQ,
    ALG_DUMMY,
    ALG_ELEMENT_LIV,
    ALG_ELEMENT_LIC,
    ALG_EXACTLY_EQ,
    ALG_EXACTLY_TRUE,
    ALG_GCC,
    ALG_LEXICOGRAPHIC_LEQ,
    ALG_MAX_EQ,
    ALG_MAX_LEQ,
    ALG_MIN_EQ,
    ALG_MIN_GEQ,
    ALG_RELATION,
) = tuple(range(18))


GET_TRIGGERS_FCTS = [
    get_triggers_and,
    get_triggers_affine_eq,
    get_triggers_affine_geq,
    get_triggers_affine_leq,
    get_triggers_alldifferent,
    get_triggers_count_eq,
    get_triggers_dummy,
    get_triggers_element_liv,
    get_triggers_element_lic,
    get_triggers_exactly_eq,
    get_triggers_exactly_true,
    get_triggers_gcc,
    get_triggers_lexicographic_leq,
    get_triggers_max_eq,
    get_triggers_max_leq,
    get_triggers_min_eq,
    get_triggers_min_geq,
    get_triggers_relation,
]

GET_COMPLEXITY_FCTS = [
    get_complexity_and,
    get_complexity_affine_eq,
    get_complexity_affine_geq,
    get_complexity_affine_leq,
    get_complexity_alldifferent,
    get_complexity_count_eq,
    get_complexity_dummy,
    get_complexity_element_liv,
    get_complexity_element_lic,
    get_complexity_exactly_eq,
    get_complexity_exactly_true,
    get_complexity_gcc,
    get_complexity_lexicographic_leq,
    get_complexity_max_eq,
    get_complexity_max_leq,
    get_complexity_min_eq,
    get_complexity_min_geq,
    get_complexity_relation,
]


COMPUTE_DOMAINS_FCTS = [
    compute_domains_and,
    compute_domains_affine_eq,
    compute_domains_affine_geq,
    compute_domains_affine_leq,
    compute_domains_alldifferent,
    compute_domains_count_eq,
    compute_domains_dummy,
    compute_domains_element_liv,
    compute_domains_element_lic,
    compute_domains_exactly_eq,
    compute_domains_exactly_true,
    compute_domains_gcc,
    compute_domains_lexicographic_leq,
    compute_domains_max_eq,
    compute_domains_max_leq,
    compute_domains_min_eq,
    compute_domains_min_geq,
    compute_domains_relation,
]

COMPUTE_DOMAIN_SIGNATURE = int64(int32[:, :], int32[:])
COMPUTE_DOMAIN_TYPE = types.FunctionType(COMPUTE_DOMAIN_SIGNATURE)
COMPUTE_DOMAINS_ADDRS = (
    np.array(build_function_address_list(COMPUTE_DOMAINS_FCTS, COMPUTE_DOMAIN_SIGNATURE))
    if not NUMBA_DISABLE_JIT
    else np.empty(0)
)


@njit(cache=True)
def pop_propagator(triggered_propagators: NDArray, not_entailed_propagators: NDArray, previous_prop_idx: int) -> int:
    for prop_idx, triggered_prop in enumerate(triggered_propagators):
        if triggered_prop and not_entailed_propagators[prop_idx] and prop_idx != previous_prop_idx:
            triggered_propagators[prop_idx] = False
            return prop_idx
    return -1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy import new_triggers


def get_complexity_element_lic(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


def get_triggers_element_lic(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domains_element_lic(domains: NDArray, parameters: NDArray) -> int:
    """
    Enforces l_i = c.
    :param domains: the domains of the variables, l is the list of the first n-1 domains, i is the last domain
    :param parameters: the parameters of the propagator, c is the first parameter
    :return: the status of the propagation (consistency, inconsistency or entailement) as an int
    """
    l = domains[:-1]
    i = domains[-1]
    # update i
    i[MIN] = max(i[MIN], 0)
    i[MAX] = min(i[MAX], len(l) - 1)
    if i[MAX] < i[MIN]:
        return PROP_INCONSISTENCY
    c = parameters[0]
    i_min = i[MIN]
    i_max = i[MAX]
    for idx in range(i[MIN], i[MAX] + 1):
        if c < l[idx, MIN] or c > l[idx, MAX]:  # no intersection
            if idx == i_min:
                i_min += 1
            elif idx == i_max:
                i_max -= 1
    if i_max < i_min:
        return PROP_INCONSISTENCY
    i[MIN] = i_min
    i[MAX] = i_max
    # when strict, update l
    # for idx in range(0, i[MIN]):
    #     if l[idx, MIN] == c:
    #         l[idx, MIN] = c + 1
    #     elif l[idx, MAX] == c:
    #         l[idx, MAX] = c - 1
    #     if l[idx, MIN] > l[idx, MAX]:
    #         return PROP_INCONSISTENCY
    # for idx in range(i[MAX] + 1, len(l)):
    #     if l[idx, MIN] == c:
    #         l[idx, MIN] = c + 1
    #     elif l[idx, MAX] == c:
    #         l[idx, MAX] = c - 1
    #     if l[idx, MIN] > l[idx, MAX]:
    #         return PROP_INCONSISTENCY
    if i_min == i_max:
        l[i_min, MIN] = l[i_max, MAX] = c
        return PROP_ENTAILMENT
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    END,
    MAX,
    MIN,
    PROBLEM_FILTERED,
    PROBLEM_INCONSISTENT,
    PROBLEM_SOLVED,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    START,
)
from nucs.numba import NUMBA_DISABLE_JIT, function_from_address
from nucs.problems.problem import Problem, is_solved
from nucs.propagators.propagators import (
    COMPUTE_DOMAIN_TYPE,
    COMPUTE_DOMAINS_ADDRS,
    COMPUTE_DOMAINS_FCTS,
    pop_propagator,
)
from nucs.statistics import (
    STATS_IDX_PROBLEM_FILTER_NB,
    STATS_IDX_PROPAGATOR_ENTAILMENT_NB,
    STATS_IDX_PROPAGATOR_FILTER_NB,
    STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_INCONSISTENCY_NB,
)


def bound_consistency_algorithm(statistics: NDArray, problem: Problem) -> int:
    """
    Applies the bound consistency algorithm.
    :param statistics: the statistics array
    :param problem: the problem
    :return: the status as an integer
    """
    return _bound_consistency_algorithm(
        statistics,
        problem.algorithms,
        problem.var_bounds,
        problem.param_bounds,
        problem.props_dom_indices,
        problem.props_dom_offsets,
        problem.props_parameters,
        problem.shr_domains_arr,
        problem.shr_domains_propagators,
        problem.triggered_propagators,
        problem.not_entailed_propagators,
        COMPUTE_DOMAINS_ADDRS,
    )


@njit(cache=True)
def _bound_consistency_algorithm(
    statistics: NDArray,
    algorithms: NDArray,
    var_bounds: NDArray,
    data_bounds: NDArray,
    props_indices: NDArray,
    props_offsets: NDArray,
    props_data: NDArray,
    shr_domains: NDArray,
    shr_domains_props: NDArray,
    triggered_props: NDArray,
    not_entailed_props: NDArray,
    compute_domains_addrs: NDArray,
) -> int:
    """
    Internal method for applying the bound consistency algorithm.
    This method only uses Numpy arrays as parameters, this permits JIT compilation.
    """
    statistics[STATS_IDX_PROBLEM_FILTER_NB] += 1
    prop_idx = -1
    while True:
        prop_idx = pop_propagator(triggered_props, not_entailed_props, prop_idx)
        if prop_idx == -1:
            return PROBLEM_SOLVED if is_solved(shr_domains) else PROBLEM_FILTERED
        statistics[STATS_IDX_PROPAGATOR_FILTER_NB] += 1
        prop_var_start = var_bounds[prop_idx, START]
        prop_var_end = var_bounds[prop_idx, END]
        prop_indices = props_indices[prop_var_start:prop_var_end]
        prop_offsets = props_offsets[prop_var_start:prop_var_end]
        prop_var_nb = prop_var_end - prop_var_start
        prop_domains = np.empty((2, prop_var_nb), dtype=np.int32).T  # trick for order=F
        np.add(shr_domains[prop_indices], prop_offsets, prop_domains)
        algorithm = algorithms[prop_idx]
        compute_domains_function = (
            COMPUTE_DOMAINS_FCTS[algorithm]
            if NUMBA_DISABLE_JIT
            else function_from_address(COMPUTE_DOMAIN_TYPE, compute_domains_addrs[algorithm])
        )
        prop_data = props_data[data_bounds[prop_idx, START] : data_bounds[prop_idx, END]]
        status = compute_domains_function(prop_domains, prop_data)
        if status == PROP_INCONSISTENCY:
            statistics[STATS_IDX_PROPAGATOR_INCONSISTENCY_NB] += 1
            return PROBLEM_INCONSISTENT
        if status == PROP_ENTAILMENT:
            not_entailed_props[prop_idx] = False
            statistics[STATS_IDX_PROPAGATOR_ENTAILMENT_NB] += 1
        shr_domains_changes = False
        for var_idx in range(prop_var_nb):
            shr_domain_idx = prop_indices[var_idx]
            prop_offset = prop_offsets[var_idx][0]
            for bound in [MIN, MAX]:
                shr_domain_bound = prop_domains[var_idx, bound] - prop_offset
                if shr_domains[shr_domain_idx, bound] != shr_domain_bound:
                    shr_domains[shr_domain_idx, bound] = shr_domain_bound
                    shr_domains_changes = True
                    np.logical_or(triggered_props, shr_domains_props[shr_domain_idx, bound], triggered_props)
        if not shr_domains_changes:  # type: ignore
            statistics[STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB] += 1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
START = 0  # index corresponding the start of a values range
END = 1  # index corresponding the end of a values range
MIN = 0  # min value of a domain
MAX = 1  # max value of a domain
PROP_INCONSISTENCY = 0  # returned by a propagator when inconsistent
PROP_CONSISTENCY = 1  # returned by a propagator when consistent
PROP_ENTAILMENT = 2  # returned by a propagator when entailed
PROBLEM_TO_FILTER = 0  # returned when the problem needs some filtering
PROBLEM_FILTERED = 1  # returned when the filtering of a problem has been completed
PROBLEM_INCONSISTENT = 2  # returned when the filtering of a problem detects an inconsistency
PROBLEM_SOLVED = 3  # returned when a problem is solved
